/*****************************************************************************/
/*****************************************************************************/
/**                                                                         **/
/**      IBM SCSI Tape & Medium Changer Device Driver / HP-UX               **/
/**            (C) COPYRIGHT IBM Corporation, 1998, 2007                    **/
/**                                                                         **/
/**          smc.h - IBM SCSI Media Changer Interface Include File          **/
/**                  Licensed Materials - Property of IBM                   **/
/**                           All Rights Reserved                           **/
/**                  Version 1.0.6   10/31/2007                             **/
/**                                                                         **/
/**    US Government Users Restricted Rights  -  Use, duplication, or       **/
/**    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.    **/
/**                                                                         **/
/*****************************************************************************/
/*****************************************************************************/

#ifndef SMC_H
#define SMC_H

#ifndef UCHAR
#define UCHAR
typedef unsigned char   uchar;
#endif

/*===========================================================================*/
/* Basic SCSI IOCTL Command Definitions and Associated Structures            */
/*===========================================================================*/

#ifndef BASIC_SCSI_IOC
#define BASIC_SCSI_IOC

/*---------------------------------------------------------------------------*/
/* Defines and data structures for INQUIRY                                   */
/*---------------------------------------------------------------------------*/

#define IOC_INQUIRY                     _IOR('S',1,inquiry_data_t)

typedef struct {
  uchar qual            : 3,            /* peripheral qualifier              */
        type            : 5;            /* device type                       */
  uchar rm              : 1,            /* removable medium                  */
        mod             : 7;            /* device type modifier              */
  uchar iso             : 2,            /* ISO version                       */
        ecma            : 3,            /* EMCA version                      */
        ansi            : 3;            /* ANSI version                      */
  uchar aen             : 1,            /* asynchronous even notification    */
        trmiop          : 1,            /* terminate I/O process message     */
                        : 2,            /* reserved                          */
        rdf             : 4;            /* response data format              */
  uchar len;                            /* additional length                 */
  uchar                 : 8;            /* reserved                          */
  uchar                 : 1,            /* reserved                          */
        encsrv          : 1,            /* enclosure service                 */
        barcod          : 1,            /* bar code scanner attached         */
        multip          : 1,            /* multi-port                        */
        mchngr          : 1,            /* medium changer mode               */
                        : 3;            /* reserved                          */
  uchar reladr          : 1,            /* relative addressing               */
        wbus32          : 1,            /* 32 bit wide data transfers        */
        wbus16          : 1,            /* 16 bit wide data transfers        */
        sync            : 1,            /* synchronous data transfers        */
        linked          : 1,            /* linked commands                   */
                        : 1,            /* reserved                          */
        cmdque          : 1,            /* command queueing                  */
        sftre           : 1;            /* soft reset                        */
  uchar vid[8];                         /* vendor ID                         */
  uchar pid[16];                        /* product ID                        */
  uchar rev[4];                         /* product revision level            */
  uchar vendor[92];                     /* vendor specific (padded to 128)   */
} inquiry_data_t;

/*---------------------------------------------------------------------------*/
/* Defines and data structures for REQUEST_SENSE                             */
/*---------------------------------------------------------------------------*/

#define IOC_REQUEST_SENSE               _IOR('T',2,sense_data_t)

typedef struct {
  uchar valid           : 1,            /* sense data is valid               */
        code            : 7;            /* error code                        */
  uchar segnum;                         /* segment number                    */
  uchar fm              : 1,            /* file mark detected                */
        eom             : 1,            /* end of medium                     */
        ili             : 1,            /* incorrect length indicator        */
                        : 1,            /* reserved                          */
        key             : 4;            /* sense key                         */
  uchar info[4];                        /* information bytes                 */
  uchar addlen;                         /* additional sense length           */
  uchar cmdinfo[4];                     /* command specific information      */
  uchar asc;                            /* additional sense code             */
  uchar ascq;                           /* additional sense code qualifier   */
  uchar fru;                            /* field replaceable unit code       */
  uchar sksv            : 1,            /* sense key specific valid          */
        cd              : 1,            /* control/data                      */
                        : 2,            /* reserved                          */
        bpv             : 1,            /* bit pointer valid                 */
        sim             : 3;            /* system information message        */
  uchar field[2];                       /* field pointer                     */
  uchar vendor[110];                    /* vendor specific (padded to 128)   */
} sense_data_t;

/*---------------------------------------------------------------------------*/
/* Defines for TEST_UNIT_READY                                               */
/*---------------------------------------------------------------------------*/

#define IOC_TEST_UNIT_READY             _IO('S',3)

/*---------------------------------------------------------------------------*/
/* Defines for RESERVE and RELEASE                                           */
/*---------------------------------------------------------------------------*/

#define IOC_RESERVE                     _IO('S',4)

#define IOC_RELEASE                     _IO('S',5)


/*---------------------------------------------------------------------------*/
/* Defines and data structures for LOG_SENSE_PAGE                            */
/*---------------------------------------------------------------------------*/

#define IOC_LOG_SENSE_PAGE              _IOWR('S',6, log_sns_pg_t)

#define MAX_LGPGDATA 250                /* The maximum data length which     */
					/* this ioctl can return, including  */
					/* the log page header.              */

typedef struct {
   uchar  page_code;                    /* page code                         */
   uchar  data[MAX_LGPGDATA];           /* log data structure                */
} log_sns_pg_t;

#define LOGSENSEPAGE 1024
#define SIOC_LOG_SENSE10_PAGE      	 _IOWR('S',12, log_sense10_page_t)

typedef struct {
   uchar page_code;			    	/* [IN] Log sense page */
   uchar subpage_code;		        /* [IN] Log sense subpage */
   uchar reserved[2];			    /* unused */
   unsigned short len;			    /* [OUT] number of valid bytes in data (log_page_header_size + page_length) */
   unsigned short parm_pointer;		/* [IN] specific parameter number at which the data begins */
   char data[LOGSENSEPAGE];	    	/* [OUT] log data */
} log_sense10_page_t;


/*
** Log Page Data Structures
**
** Portions of the log_sns_pg_t data[] field may be
** cast to these structures when interpreting the mode data which
** is returned by IOC_LOG_SENSE_PAGE.
**
** Here is the generalized structure of a log page:
**
** Log Page
**    Log Page Header
**       Page Code
**       Page Length
**    Log Parameter(s) (One or more may exist)
**       Parameter Code
**       Control Byte
**       Parameter Length
**       Parameter Value
*/

typedef struct {
   uchar page_code;                     /* page code                         */
   uchar reserved_1;                    /* reserved                          */
   uchar len[2];                        /* page length                       */
} log_page_hdr_t;

typedef struct {
   uchar parm_code[2];                  /* parameter code                    */
   uchar ctl_byte;                      /* control byte                      */
   uchar parm_len;                      /* parameter length                  */
} log_parm_hdr_t;

/*---------------------------------------------------------------------------*/

#if 0
#define MAXLOGPARMS \                  /* Maximum page parameters length.    */
   (MAX_LGPGDATA - sizeof(log_page_hdr_t))
#endif

#if 0
typedef struct {
   log_page_hdr_t header;
   char list[MAXLOGPARMS];
} log_pg_00_t;
#endif

typedef struct {
   log_page_hdr_t page_hdr;
   log_parm_hdr_t parm_hdr;
   uchar alert_value;                  /* tape alert flag                    */
} log_pg_2E_t;

/*SIM/MIM page*/
typedef struct {
   log_page_hdr_t page_hdr;
   log_parm_hdr_t parm_hdr;
   uchar indicator;                    /* SIM MIM indicator                  */
   uchar res1[11];                     /* reserved                           */
   uchar mcode[2];                     /* message code                       */
   uchar res2[2];                      /* reserved                           */
   uchar exception_mcode;              /* exception message code             */
   uchar service_mcode;                /* service message code               */
   uchar severity_code;                /* message severity code              */
   uchar res3;                         /* reserved                           */
   uchar exception_data[2];            /* exception data                     */
   uchar fru_id[3];                    /* FRU id in SIM or FSC in MIM        */
   uchar fsc1[4];                      /* first FSC                          */
   uchar fsc2[4];                      /* last FSC                           */
   uchar product_id[4];                /* product ID                         */
   uchar product_ident[18];            /* product identifier                 */
   uchar device_type[8];               /* device type and model number       */
} log_pg_31_t;


/*---------------------------------------------------------------------------*/
/* Defines and data structures for MODE_SENSE and MODE_PAGE                  */
/*---------------------------------------------------------------------------*/

#define IOC_MODE_SENSE                 _IOWR('S',7, mode_sns_t)

#define MAX_MSDATA 253                 /* The maximum data length which this */
				       /* ioctl can return, including        */
    				       /* headers and block descriptors.     */

#define MODESNS_10_CMD 0x5A            /* SCSI cmd code for 10-byte version  */
        			       /* of the command                     */

#define MODESNS_6_CMD  0x1A            /* SCSI cmd code for 6-byte version   */
    				       /* of the command                     */

typedef struct {
   uchar page_code;                    /* Page Code:  Set this field with    */
				       /*    the desired mode page number    */
				       /*    before issuing the ioctl.       */
   uchar cmd_code;                     /* SCSI Command Code:  Upon return,   */
				       /*    this field is set with the      */
				       /*    SCSI command code to which      */
				       /*    the device responded.           */
				       /*    x'5A' = Mode Sense (10)         */
				       /*    x'1A' = Mode Sense (6)          */
   uchar data[MAX_MSDATA];             /* Mode Parameter List:  Upon return, */
                     	               /*    this field contains the mode    */
				       /*    parameters list, up to the max  */
				       /*    length supported by the ioctl.  */
} mode_sns_t;

/*
** Mode Parameter List Data Structures
**
** Portions of the mode_sns_t data[] field may be cast to
** these structures when interpreting the mode parameter list
** returned by IOC_MODE_SENSE.
**
** Here is the generalized structure of a mode parmameter list:
**
** Mode Parameter List
**    Mode Parameter Header (One of the two possible versions)
**       Data Length
**       Medium Type
**       Device-Specific Parameter
**       Block Descriptor Length
**    Block Descriptor(s) (Zero or more may be present)
**       Density Code
**       Block Count
**       Block Length
**    Mode Page(s) (One or more may be present)
**       Page Code
**       Page Length
**       Mode Parameters
**
** The Mode Parameter Header begins at byte 0 of the
** mode_sns_t data[] field.
*/

/* Mode Parameter Header for 10-byte version of Mode Sense.                  */
typedef struct {
   uchar data_len[2];                  /* Mode Data Length which follows     */
                  		       /* this field                         */
   uchar medium_type;                  /* Medium Type                        */
   uchar wrt_prot        : 1,          /* Write Protect                      */
         buf_mode        : 3,          /* Buffered Mode                      */
         speed           : 4;          /* Speed                              */
   uchar reserved1[2];                 /* Reserved                           */
   uchar blk_dsc_len[2];               /* Block Descriptor Length            */
} mode_hdr10_t;

/* Mode Parameter Header for 6-byte version of Mode Sense.                   */
typedef struct {
   uchar data_len;                     /* Mode Data Length which follows     */
				       /*    this field                      */
   uchar medium_type;                  /* Medium Type                        */
   uchar wrt_prot        : 1,          /* Write Protect                      */
         buf_mode        : 3,          /* Buffered Mode                      */
         speed           : 4;          /* Speed                              */
   uchar blk_dsc_len;                  /* Block Descriptor Length            */
} mode_hdr6_t;

/* If blk_dsc_len is non-zero, then one or more block descriptors exists.    */
/* Number of block descriptors = blk_desc_len / sizeof(blkdsc_t)             */
typedef struct {
   uchar density_code;                 /* Density Code                       */
   uchar blk_count[3];                 /* Number of Blocks                   */
   uchar reserved1;                    /* Reserved                           */
   uchar blk_len[3];                   /* Block Length                       */
} blkdsc_t ;

/*
** The maximum length of mode page data which can be returned by
** IOC_MODE_SENSE depends on whether the six-byte or ten-byte version
** of Mode Sense is used by the tape device, and whether or not the
** returned mode parameter list includes one or more block descriptors.
**
** For any given call to IOC_MODE_SENSE, the maximum length of
** returned mode data can be calculated as:
**
**    MAX_MSDATA
**       - <size_of_mode_parameter_header_structure>
**       - <block_descriptor_length_in_mode_parameter_header>
**
** This maximum holds, regardless of the Page Length value indicated in the
** mode page itself.
**
** The first two bytes of the mode page data may be interpreted using the
** following structure.  All remaining bytes must be interpreted according
** to the description of each individual mode page in the appropriate
** hardware reference.
*/
typedef struct {
   uchar ps            : 1,            /* Parameter Saveable                 */
         reserved_1    : 1,            /* reserved                           */
         page_code     : 6;            /* Page Code                          */
   uchar page_len;                     /* Page Length of following data      */
} modep_prefix_t;


/*---------------------------------------------------------------------------*/
/* Defines and data structures for INQUIRY_PAGE                              */
/*---------------------------------------------------------------------------*/

#define IOC_INQUIRY_PAGE               _IOWR('S',8, inquiry_page_t)

#define MAX_IPDATA   253               /* max data xfer for inquiry pg ioctl */

typedef struct {
   uchar page_code;                     /* page code                          */
   uchar data[MAX_IPDATA];              /* inquiry parameter List             */
} inquiry_page_t;

typedef struct {
   uchar periph_qual   : 3,            /* peripheral qualifier               */
         periph_type   : 5;            /* peripheral device type             */
   uchar page_code;                    /* page code                          */
   uchar reserved_1;                   /* reserved                           */
   uchar page_len;                     /* page length                        */
   uchar serial[12];                   /* serial number                      */
} inq_pg_80_t;

typedef struct {
   uchar periph_qual   : 3,            /* peripheral qualifier               */
         periph_type   : 5;            /* peripheral device type             */
   uchar page_code;                    /* page code                          */
   uchar reserved_1;                   /* reserved                           */
   uchar page_len;                     /* page length                        */
   uchar vendor[2];                    /* vendor-specific world wide name    */
   uchar port;                         /* port node ID                       */
} inq_pg_83_t;


/*---------------------------------------------------------------------------*/
/* Defines for PREVENT and ALLOW_MEDIUM_REMOVAL                              */
/*---------------------------------------------------------------------------*/

#define IOC_PREVENT_MEDIUM_REMOVAL      _IO('S', 9)

#define IOC_ALLOW_MEDIUM_REMOVAL        _IO('S', 10)


/*---------------------------------------------------------------------------*/

#endif  /* BASIC_SCSI_IOC */

/*===========================================================================*/
/* SCSI Medium Changer IOCTL Command Definitions and Associated Structures   */
/*===========================================================================*/

/*---------------------------------------------------------------------------*/
/* Defines and data structures for ELEMENT_INFO                              */
/*---------------------------------------------------------------------------*/

#define SMCIOC_ELEMENT_INFO             _IOR('C',3,element_info_t)

typedef struct {
  ushort robot_address;                 /* medium transport element address  */
  ushort robot_count;                   /* number medium transport elements  */
  ushort cell_address;                  /* medium storage element address    */
  ushort cell_count;                    /* number medium storage elements    */
  ushort port_address;                  /* import-export element address     */
  ushort port_count;                    /* number import-export elements     */
  ushort drive_address;                 /* data transfer element address     */
  ushort drive_count;                   /* number data transfer elements     */
} element_info_t;

/*---------------------------------------------------------------------------*/
/* Defines and data structures for MOVE_MEDIUM                               */
/*---------------------------------------------------------------------------*/

#define SMCIOC_MOVE_MEDIUM              _IOW('C',4,move_medium_t)

typedef struct {
  ushort robot;                         /* robot address                     */
  ushort source;                        /* move from location                */
  ushort destination;                   /* move to location                  */
  uchar  invert;                        /* invert medium before insertion    */
} move_medium_t;

#define NO_FLIP         0               /* do not invert medium              */
#define FLIP            1               /* invert medium                     */


/*---------------------------------------------------------------------------*/
/* Defines and data structures for POSITION_TO_ELEMENT                       */
/*---------------------------------------------------------------------------*/

#define SMCIOC_POS_TO_ELEM              _IOW('C',5,pos_to_elem_t)

typedef struct {
  ushort robot;                         /* robot address                     */
  ushort destination;                   /* move to location                  */
  uchar  invert;                        /* invert medium before insertion    */
} pos_to_elem_t;


/*---------------------------------------------------------------------------*/
/* Defines for AUDIT                                                         */
/*---------------------------------------------------------------------------*/

#define SMCIOC_AUDIT                    _IO('C',6)


/*---------------------------------------------------------------------------*/
/* Defines and data structures for INVENTORY                                 */
/*---------------------------------------------------------------------------*/

#define SMCIOC_INVENTORY                _IOW('C',7,inventory_t)

#if defined(__LP64__) && defined (_KERNEL)
#define SMCIOC_INVENTORY32              _IOW('C',7,inventory_t32)
#endif

typedef struct {
  ushort address;                       /* element address                   */
  uchar                 : 2,            /* reserved                          */
        inenab          : 1,            /* medium in robot scope             */
        exenab          : 1,            /* medium not in robot scope         */
        access          : 1,            /* robot access allowed              */
        except          : 1,            /* abnormal element state            */
        impexp          : 1,            /* medium imported or exported       */
        full            : 1;            /* element contains medium           */
  uchar                 : 8;            /* reserved                          */
  uchar asc;                            /* additional sense code             */
  uchar ascq;                           /* additional sense code qualifier   */
  uchar notbus          : 1,            /* element not on same bus as robot  */
                        : 1,            /* reserved                          */
        idvalid         : 1,            /* scsi bus id valid                 */
        luvalid         : 1,            /* logical unit valid                */
                        : 1,            /* reserved                          */
        lun             : 3;            /* logical unit                      */
  uchar scsi;                           /* scsi bus id                       */
  uchar                 : 8;            /* reserved                          */
  uchar svalid          : 1,            /* element address valid             */
        invert          : 1,            /* medium inverted                   */
                        : 6;            /* reserved                          */
  ushort source;                        /* source storage element address    */
  uchar volume[36];                     /* primary volume tag                */
  uchar vendor[80];                     /* vendor specific (padded to 128)   */
} element_status_t;

typedef struct {
  ptr32_t robot_status;       /* medium transport element pages    */
  ptr32_t cell_status;        /* medium storage element pages      */
  ptr32_t port_status;        /* import-export element pages       */
  ptr32_t drive_status;       /* data transfer element pages       */
} inventory_t32;

typedef struct {
  element_status_t *robot_status;       /* medium transport element pages    */
  element_status_t *cell_status;        /* medium storage element pages      */
  element_status_t *port_status;        /* import-export element pages       */
  element_status_t *drive_status;       /* data transfer element pages       */
} inventory_t;


/*---------------------------------------------------------------------------*/
/* Defines for LOCK_DOOR and UNLOCK_DOOR                                     */
/*---------------------------------------------------------------------------*/

#define SMCIOC_LOCK_DOOR                _IOW('C',8,lock_door_t)

typedef uchar lock_door_t;

#define SMC_UNLOCK       0               /* unlock door                   */

#define SMC_LOCK         1               /* lock door                     */


/*---------------------------------------------------------------------------*/
/* Defines and data structures for READ_ELEMENT_DEVIDS (DVCID = 1)           */
/*---------------------------------------------------------------------------*/

#define SMCIOC_READ_ELEMENT_DEVIDS      _IOW('C',9,read_element_devids_t)

#if defined(__LP64__) && defined (_KERNEL)
#define SMCIOC_READ_ELEMENT_DEVIDS32    _IOW('C',9,read_element_devids_t32)
#endif

typedef struct {
    ushort address;             /* element address */
    uchar          :4,          /* reserved */
           access  :1,          /* robot access allowed */
           except  :1,          /* abnormal element state */
                   :1,          /* reserved */
           full    :1;          /* element contains medium */
    uchar  resvd1;              /* reserved */
    uchar  asc;                 /* additional sense code */
    uchar  ascq;                /* additional sense code qualifier */
    uchar  notbus  :1,          /* element not on same bus as robot */
                   :1,          /* reserved */
           idvalid :1,          /* element address valid */
           luvalid :1,          /* logical unit valid */
                   :1,          /* reserved */
           lun     :3;          /* logical unit number */
    uchar  scsi;                /* scsi bus address */
    uchar  resvd2;              /* reserved */
    uchar  svalid  :1,          /* element address valid */
           invert  :1,          /* medium inverted */
                   :6;          /* reserved */
    ushort source;              /* source storage element address */
    uchar          :4,          /* reserved */
           code_set:4;          /* code set X'2' is all ASCII identifier */
    uchar          :4,          /* reserved */
           id_type :4;          /* identifier type */
    uchar  resvd3;              /* reserved */
    uchar  id_len;              /* identifier length */
    uchar  dev_id[36];          /* device identification with serial number */
} element_devid_t;

typedef struct {
    ushort element_address;               /* starting element address */
    ushort number_elements;               /* number of elements       */
    element_devid_t *drive_devid;         /* data transfer element pages */
} read_element_devids_t;

typedef struct {
    ushort element_address;               /* starting element address */
    ushort number_elements;               /* number of elements       */
    ptr32_t drive_devid;                  /* data transfer element pages */
} read_element_devids_t32;


/*---------------------------------------------------------------------------*/
/* Defines and data structures for EXCHANGE_MEDIUM                           */
/*---------------------------------------------------------------------------*/

#define SMCIOC_EXCHANGE_MEDIUM          _IOW('C',10,exchange_medium_t)

typedef struct {
    ushort robot;               	/* robot address		 */
    ushort source;              	/* move from location		 */
    ushort destination1;        	/* move to location		 */
    ushort destination2;        	/* move to location		 */
    uchar invert1;              	/* invert before placement into destination 1 */
    uchar invert2;              	/* invert before placement into destination 2 */
} exchange_medium_t;


/*---------------------------------------------------------------------------*/
/* Defines and data structures for INIT_ELEM_STAT_RANGE                      */
/*---------------------------------------------------------------------------*/

#define SMCIOC_INIT_ELEM_STAT_RANGE     _IOW('C',11,element_range_t)

typedef struct {
    ushort element_address;     	/* starting element address	 */
    ushort number_elements;     	/* number of elements      	 */
} element_range_t;


/*---------------------------------------------------------------------------*/
/* Defines and data structures for SMCIOC_READ_CARTRIDGE_LOCATION            */
/*---------------------------------------------------------------------------*/

#define SMCIOC_READ_CARTRIDGE_LOCATION   _IOWR('C',12,read_cartridge_location_t)
#if defined(__LP64__) && defined (_KERNEL)
#define SMCIOC_READ_CARTRIDGE_LOCATION32 _IOWR('C',12,read_cartridge_location_t32)
#endif


typedef struct
{
    ushort address;                            /* element address          */
    uchar  :4,                                 /* reserved                 */
           access:1,                           /* robot access allowed     */
           except:1,                           /* abnormal element state   */
                 :1,                           /* reserved                 */
           full:1;                             /* element contains medium  */
    uchar  resvd1;                             /* reserved                 */
    uchar  asc;                                /* additional sense code    */
    uchar  ascq;                               /* additional sense code qualifier */
    uchar  resvd2[3];                          /* reserved                 */
    uchar  svalid:1,                           /* element address valid    */
           invert:1,                           /* medium inverted          */
           :6;                                 /* reserved                 */
    ushort source;                             /* source storage elem addr */
    uchar  volume[36];                         /* primary volume tag       */
    uchar  :4,                                 /* reserved                 */
           code_set:4;                         /* code set                 */
    uchar  :4,                                 /* reserved                 */
           ident_type:4;                       /* identifier type          */
    uchar  resvd3;                             /* reserved                 */
    uchar  ident_len;                          /* identifier length        */
    uchar  identifier[24];                     /* slot identification      */
} cartridge_location_data_t;

typedef struct
{
    ushort element_address;                    /* starting element address */
    ushort number_elements;                    /* number of elements       */
    cartridge_location_data_t *data;           /* storage element pages    */
    char reserved[8];                          /* reserved                 */
} read_cartridge_location_t;

typedef struct
{
    ushort element_address;                    /* starting element address */
    ushort number_elements;                    /* number of elements       */
    ptr32_t  data;                             /* storage element pages    */
    char reserved[8];                          /* reserved                 */
} read_cartridge_location_t32;


/*---------------------------------------------------------------------------*/

#define SMCIOC_CHANGE_DEFINITION           _IOW('C', 101,int)

#endif  /* SMC_H */


